---------------------------------------------------
--
-- Flash Controller
--
-- Copyleft 2003 Hkan Kvist & Linus Walleij
--
---------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use work.controller_types.all;

package flashctl_pck is
  component flashctl
  port ( clk			: in std_logic;
	 rst			: in std_logic;
	 cpu_clk		: out std_logic; 
	 -- Interface to Flash
	 flashi			: in flash_out_type;
	 flasho                 : out flash_in_type;
	 -- For using Flash as ROM
	 flash_rom_data_out	: out std_logic_vector(7 downto 0);
	 flash_rom_re		: in std_logic;
	 flash_rom_address_in	: in std_logic_vector(15 downto 0);
	 -- MP3 streaming registers
	 play_start_address     : in std_logic_vector (31 downto 0);
	 play_end_address       : in std_logic_vector (31 downto 0);
	 play                   : in std_logic;
	 stop                   : in std_logic;
	 stopped                : out std_logic;
	 -- Interface to MP3 chip
         mp3i			: in  mp3_out_type;
	 mp3o			: out mp3_in_type
       );
  end component;
end;

-------------------------------------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
library work;
use work.flashctl_pck.all;
use work.controller_types.all;
use std.textio.all;

entity flashctl is
  port ( clk			: in std_logic;
	 rst			: in std_logic;
	 cpu_clk		: out std_logic;
	 -- Interface to Flash
	 flashi			: in flash_out_type;
	 flasho                 : out flash_in_type;
	 -- For using Flash as ROM
	 flash_rom_data_out	: out std_logic_vector(7 downto 0);
	 flash_rom_re		: in std_logic;
	 flash_rom_address_in	: in std_logic_vector(15 downto 0);
	 -- MP3 streaming registers
	 play_start_address     : in std_logic_vector (31 downto 0);
	 play_end_address       : in std_logic_vector (31 downto 0);
	 play                   : in std_logic;
	 stop                   : in std_logic;
	 stopped                : out std_logic;
	 -- Interface to MP3 chip
         mp3i			: in  mp3_out_type;
	 mp3o			: out mp3_in_type
       );
end;

-------------------------------------------------------------------------------
-- HAPPY ARCHITECTURE OF THIS FLASH CONTROLLER
-------------------------------------------------------------------------------

architecture KNAST of flashctl is

-------------------------------------------------------------------------------
-- mr printhelper
    procedure print(s : string) is
 	variable L1 : line;
      begin
	  L1:= new string'(s);
 	writeline(output,L1);
     end;

    procedure printbyte(b : std_logic_vector(7 downto 0)) is
	variable L1 : line;
    begin
        write(L1, to_bitvector(b));
	writeline(output,L1);
    end;

    procedure printaddress(a : std_logic_vector(19 downto 0)) is
	variable L1 : line;
    begin
        write(L1, to_bitvector(a));
	writeline(output,L1);
    end;
-------------------------------------------------------------------------------

    constant DELAY_CYCLES_WRITE : std_logic_vector(3 downto 0) := "0010";

--    type state_type is ( IDLE, WAIT_CPU_ADDRESS, WAIT_CPU, WAIT_CPU2, READOUT_CPU, WAIT_MP3, WAIT_MP32, READOUT_MP3 );
    type state_type is ( IDLE, WAIT_CPU, WAIT_CPU2, READOUT_CPU, WAIT_MP3, WAIT_MP32, READOUT_MP3 );
    type playstate_type is ( STOPMODE, PLAYMODE, PAUSEMODE );
    
    type output_buffer_type is record
	     flipper : std_logic;
	     state : state_type;
	     playstate : playstate_type;
	     play_address : std_logic_vector (FLASH_ADDR_BITS-1 downto 0);
	     -- Buffered outputs
	     flash_wdata : std_logic_vector (FLASH_DATA_BITS-1 downto 0);
	     flash_we : std_logic;
	     flash_oe : std_logic;
	     stopped : std_logic;
	     -- To MP3 chip
	     mp3byte : std_logic_vector (7 downto 0);
	     bitcount : std_logic_vector (3 downto 0);
	     mp3o : mp3_in_type;
	     -- Requested ROM address
	     req_address : std_logic_vector(FLASH_ADDR_BITS-1 downto 0);
	     req_byte : std_logic_vector(7 downto 0);
	     rom_served : std_logic;
    end record;

    -- The buffer pairs
    signal buff : output_buffer_type;
    signal buff_in : output_buffer_type;

    -- Signals between address buffer and state machine
    signal rom_memory_served : std_logic;

begin

    mr_buffer : process(clk)
    begin
	-- This inhibits the CPU while retrieving from Flash if needed
	-- DANGEROUSLY ASYNCHRONOUS but what else can you do.
-- 	if (flash_rom_re = '1' and rom_memory_served = '0') then
--  	    cpu_clk <= '0';
-- 	else
-- 	    cpu_clk <= clk;
-- 	end if;
	cpu_clk <= ((not flash_rom_re) or (rom_memory_served)) and clk;
	if (rising_edge(clk)) then
	    -- Copy outputs to ports
	    flasho.wdata <= buff_in.flash_wdata;
	    flasho.we <= buff_in.flash_we;
	    flasho.oe <= buff_in.flash_oe;
	    flasho.en <= '0'; -- Chip always enabled
	    mp3o <= buff_in.mp3o;
	    stopped <= buff_in.stopped;
	    flasho.addr <= buff_in.req_address;
	    flash_rom_data_out <= buff_in.req_byte;
	    rom_memory_served <= buff_in.rom_served;
	    -- Iterate
	    buff <= buff_in;
	    -- Only strobe reset and valid to MP3 chip
	    buff.mp3o.rst <= '1';
	    buff.mp3o.valid <= '0';
	    buff.rom_served <= '0';
	end if;
    end process;

    ---------------------------------------------------------------------------
    -- FLASH Process
    ---------------------------------------------------------------------------
    process(rst, buff)
	variable b : output_buffer_type;
    begin
      b := buff;
      -- Active low reset signal
      if (rst = '0') then
	  b.flipper := '0';
	  b.state := IDLE;
	  -- Init flash state
	  b.flash_we := '0';
	  b.flash_oe := '0';
	  -- Set stopped mode
	  b.playstate := STOPMODE;
	  b.stopped := '1';
	  b.bitcount := (others => '0');
	  -- b.req_byte := (others => '0');
	  -- Reset MP3 chip
	  b.mp3o.rst := '0';
	  b.mp3o.data := '0';
	  b.mp3o.valid := '0';
	  -- Byte requesting
	  b.rom_served := '0';
      else
	  -- Update play state
	  b.flipper := not b.flipper;
	  case b.playstate is
	      when STOPMODE =>
		  if (play = '1') then
		      print("Vi stller oss i playmode frn stopmode (port)");
		      b.playstate := PLAYMODE;
		      b.stopped := '0';
		      -- Set up adress counter
		      b.play_address := play_start_address(FLASH_ADDR_BITS-1 downto 0);
		  end if;
	      when PLAYMODE =>
		  if (stop = '1') then
		      print("Vi stller oss i stopmode frn playmode (port)");
		      b.playstate := STOPMODE;
		      b.stopped := '1';
		      b.mp3o.rst := '0';  -- Reset MP3 chip
		      b.mp3o.valid := '0';
		      b.bitcount := "0000";  -- Make sure nothing get sent
					     -- from buffer
		  elsif (play = '1') then
		      print("Vi stller oss i pausemode frn playmode (port)");
		      b.playstate := PAUSEMODE;
		      b.mp3o.rst := '0';  -- Reset MP3 chip
		      b.mp3o.valid := '0';
		      b.bitcount := "0000";  -- Make sure nothing get sent
					     -- from buffer
		  end if;
	      when PAUSEMODE =>
		  if (stop = '1') then
		      print("Vi stller oss i stopmode frn pausemode (port)");
		      b.playstate := STOPMODE;
		      b.stopped := '1';
		  elsif (play = '1') then
		      print("Vi stller oss i playmode frn pausemode (port)");
		      b.playstate := PLAYMODE;
		  end if;
	  end case;

	  -- Update Flash fetch states
	  case b.state is
	      when IDLE =>
		  -- Prioritize MP3 Chip requests
		  if ((mp3i.req = '1') and (b.bitcount = "0000") and (b.playstate = PLAYMODE)) then
		      -- Stop when last address is reached
		      if (b.play_address = play_end_address(FLASH_ADDR_BITS-1 downto 0)) then
			  print("Vi stller oss i stopmode frn playmode (end of song)");
			  b.playstate := STOPMODE;
			  b.stopped := '1';
			  b.mp3o.rst := '0';
			  b.mp3o.valid := '0';
			  -- Do not change state
		      else
			  -- Time to fetch next byte from Flash
			  b.req_address := b.play_address(FLASH_ADDR_BITS-1 downto 0);
			  b.state := WAIT_MP3;
			  b.play_address := b.play_address + 1;
		      end if;
		  -- CPU requests a part of the ROM memory
		  elsif (flash_rom_re = '1' and rom_memory_served = '0') then
		      b.req_address(15 downto 0) := flash_rom_address_in;
		      b.req_address(FLASH_ADDR_BITS-1 downto 16) := (others => '0');
		      print("Requested address:");
		      printaddress(b.req_address);
		      b.state := WAIT_CPU;
		  end if;
	      when WAIT_CPU =>
		  b.state := WAIT_CPU2;
	      when WAIT_CPU2 =>
		  b.state := READOUT_CPU;
	      when WAIT_MP3 =>
		  b.state := WAIT_MP32;
	      when WAIT_MP32 =>
		  b.state := READOUT_MP3;
	      when READOUT_CPU =>
		  b.req_byte := flashi.rdata;
		  -- print("Read out byte:");
		  -- printbyte(b.req_byte);
		  b.rom_served := '1';
		  b.state := IDLE;
	      when READOUT_MP3 =>
		  b.mp3byte := flashi.rdata;
		  -- Print it
		  printbyte(b.mp3byte);
		  b.bitcount := "1000";
		  -- This routine need to be doubled to assure the CPU is correctly inhibited.
		  if (flash_rom_re = '1' and rom_memory_served = '0') then
		      b.req_address(15 downto 0) := flash_rom_address_in;
		      b.req_address(FLASH_ADDR_BITS-1 downto 16) := (others => '0');
		      print("Requested address:");
		      printaddress(b.req_address);
		      b.state := WAIT_CPU;
		  else
		      b.state := IDLE;
		  end if;
	  end case;
	  
	  -- Throw bytes at mp3chip if read out
	  if (b.bitcount /= "0000") and (mp3i.req = '1') then
	      b.mp3o.data := b.mp3byte(7);
	      b.mp3o.valid := '1';
	      b.mp3byte := shl(b.mp3byte, "1");  -- Shift
	      b.bitcount := b.bitcount - 1;
	  end if;
	  
      end if;
      buff_in <= b;
    end process;
end;

