-------------------------------------------------------------------------------
-- CONTROLLER
-- (L) Copyleft 2003 Linus Walleij and Hkan Kvist
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;
library work;
use work.controller_types.all;
use work.uart_pck.all;
use work.mmu_pck.all;
use work.free_6502.all;
use work.mr_fnord_clock.all;
use work.dacuart_pck.all;
use work.ram.all;

-------------------------------------------------------------------------------
library std;
use std.textio.all;

entity controller is
    port( clk     : in  std_logic;           -- system clock 12MHz
          rst     : in  std_logic;           -- system reset
          flashi  : in  flash_out_type;      -- address of the external FLASH
          flasho  : out flash_in_type;       -- data from the external FLASH
          buttoni : in  button_type;         -- play, stop etc
          mp3i    : in  mp3_out_type;        -- to MP3 chip
          mp3o    : out mp3_in_type;         -- from MP3 chip
	  spi     : out spi_type;            -- D/A control (data)
	  rxd     : in  std_logic;           -- serial input (from PC)
          txd     : out std_logic            -- serial output (display)
         );
end;

architecture structural of controller is

  -- CPU ports and stuff
  signal cpu_addr : std_logic_vector (15 downto 0);
  signal irq      : std_logic;
  signal nmi      : std_logic;
  signal halt	  : std_logic;
  signal cpu_in   : std_logic_vector (7 downto 0);
  signal cpu_out  : std_logic_vector (7 downto 0);
  signal dout_oe  : std_logic;
  signal cpu_we   : std_logic;
  signal cpu_re   : std_logic;
  signal sync     : std_logic;
  signal rstinv   : std_logic;
  -- RAM circuit
  signal ram_data_in : std_logic_vector (7 downto 0);
  signal ram_data_out : std_logic_vector (7 downto 0);
  signal ram_we : std_logic;
  signal ram_we_inv : std_logic;
  -- LCD ports and stuff
  signal uarti    : uart_ctrl_in_type;
  signal uarto    : uart_ctrl_out_type;
  -- Mr clock and stuff
  signal clock_rst     : std_logic;
  signal clock_hold    : std_logic;
  signal clock_read    : std_logic;
  signal clock_seconds : std_logic_vector(7 downto 0);
  signal clock_minutes : std_logic_vector(7 downto 0);
  signal clock_hours   : std_logic_vector(7 downto 0);
  -- dacuart and stuff
  signal daco : spi_type;  		-- is this really needed?
  signal dac_data_byteh : std_logic_vector(7 downto 0);
  signal dac_data_bytel : std_logic_vector(7 downto 0);
  signal dac_send_data  : std_logic;
  signal dac_data_req   : std_logic;
begin

  ---------------------------------
  -- COMPONENTS ON THIS CONTROLLER
  ---------------------------------

  mr_clock: clock
      -- Change at syntehtization.
      generic map( MASTER_CLOCK => SYSTEM_CLOCK)
      -- generic map( MASTER_CLOCK => 8_000 )  -- for debugging
      port map (clk => clk,
		rst => rst,
		zero => clock_rst,
		hold => clock_hold,
		read => clock_read,
		seconds => clock_seconds,
		minutes => clock_minutes,
		hours => clock_hours
		);
    
  mr_uart: uart
    generic map ( MASTER_CLOCK => SYSTEM_CLOCK, BAUD_RATE_OUT => BAUD_RATE_LCD, BAUD_RATE_IN => BAUD_RATE_PC )
    port map ( clk   => clk,
               rst   => rst,
               rxd   => rxd,
               txd   => txd,
               uarti => uarti,
               uarto => uarto
              );

  mr_mmu : mmu
      port map (clk => clk,
		halt => halt,
		rst => rst,
		cpu_addr => cpu_addr,
                cpu_in => cpu_in,
                cpu_out => cpu_out,
                cpu_we => cpu_we,
                cpu_re => cpu_re,
		ram_data_in => ram_data_in,
		ram_data_out => ram_data_out,
		ram_we => ram_we,
		uarti => uarti,
		uarto => uarto,
		flasho => flasho,
		flashi => flashi,
		button_play => buttoni.play,
		button_stop => buttoni.stop,
		button_prev => buttoni.prev,
		button_next => buttoni.nxt,
		clock_rst => clock_rst,
		clock_hold => clock_hold,
		clock_read => clock_read,
		clock_seconds => clock_seconds,
		clock_minutes => clock_minutes,
		clock_hours => clock_hours,
		mp3i => mp3i,
		mp3o => mp3o,
		dac_data_byteh => dac_data_byteh,
		dac_data_bytel => dac_data_bytel,
		dac_send_data => dac_send_data,
		dac_data_req => dac_data_req
		);

  mr_cpu : core_6502
      port map ( clk => clk,
		 reset => rstinv,
		 irq_in => irq,
		 nmi_in => nmi,
		 addr_pin  => cpu_addr,
		 din => cpu_in,
		 dout => cpu_out,
		 dout_oe => dout_oe,
		 we_pin => cpu_we,
		 rd_pin => cpu_re,
		 sync => sync,
		 halt => halt
	       );

  mr_ram : mp3ram
      -- Remember to map adress line immediately to RAM
      generic map (BYTES => 1024)
      port map ( addr => cpu_addr(9 downto 0),
                 clk => clk,
                 data_in => ram_data_in,
                 data_out => ram_data_out,
                 we => ram_we_inv);
  
  mr_dacuart : dacuart
      port map (clk => clk,
		rst => rst,
		dacbytes(15 downto 8) => dac_data_byteh,
	        dacbytes(7 downto 0) => dac_data_bytel,
	        takethisdac => dac_send_data,
		requestdac => dac_data_req,
	        daco => spi
		);

  -- Some standard signals
  ram_we_inv <= not ram_we; -- This must be inverted
  rstinv <= not rst;
  irq <= '0';
  nmi <= '0';
  dout_oe <= '1';
end;
