---------------------------------------------------
--
-- Simple UART controller template
-- for arbitrary baud rates
--
-- Assumes LSBit is sent first
--
-- Copyleft 2003 Thomas Lenart tlt@es.lth.se
-- Copyleft 2003 Hkan Kvist & Linus Walleij
--
---------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

package uart_pck is

  type uart_ctrl_in_type is record
    data    : std_logic_vector(7 downto 0);
    send    : std_logic;
    ack_rcv : std_logic;
  end record;

  type uart_ctrl_out_type is record
    data    : std_logic_vector(7 downto 0);
    rxready : std_logic;
    txempty : std_logic; 
  end record;

  component uart
    generic ( MASTER_CLOCK : integer;
	      BAUD_RATE_OUT: integer;
	      BAUD_RATE_IN: integer
            );
    port ( clk   : in  std_logic;
           rst   : in  std_logic;
           rxd   : in  std_logic;
           txd   : out std_logic;
           uarti : in  uart_ctrl_in_type;
           uarto : out uart_ctrl_out_type
         );
  end component;

end;


-------------------------------------------------------------------------------
-- Entities
-------------------------------------------------------------------------------


library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
library work;
use work.uart_pck.all;
-- use std.textio.all;



entity uart is
  generic ( MASTER_CLOCK   : integer := 12_000_000;
            BAUD_RATE_OUT  : integer := 9600;
	    BAUD_RATE_IN   : integer := 115200
	    );
  port ( clk   : in  std_logic;
         rst   : in  std_logic;
         rxd   : in  std_logic;
         txd   : out std_logic;
         uarti : in  uart_ctrl_in_type;
         uarto : out uart_ctrl_out_type
       );
end;


-------------------------------------------------------------------------------
-- HAPPY ARCHITECTURE OF THIS UART
-------------------------------------------------------------------------------

architecture FNORD of uart is

-------------------------------------------------------------------------------
-- mr printhelper
--    procedure print(s : string) is
--	variable L1 : line;
--    begin
--	L1:= new string'(s);
--	writeline(output,L1);
--    end;
-------------------------------------------------------------------------------
  
    -- Clockfactor to count a cycle on peripheral speed.
    constant MR_CLOCKFACTOR_OUT : std_logic_vector(11 downto 0) := CONV_STD_LOGIC_VECTOR( (MASTER_CLOCK / BAUD_RATE_OUT), 12) ;
    constant MR_CLOCKFACTOR_IN : std_logic_vector(11 downto 0) := CONV_STD_LOGIC_VECTOR( (MASTER_CLOCK / BAUD_RATE_IN), 12) ;
    
    type state_type is ( IDLE, STARTBIT, DATA, STOPBIT, ACK );

    type send_reg_type is record
	     -- Internal state variables
	     state : state_type;
	     data : std_logic_vector(7 downto 0);
	     mr_factor : std_logic_vector(11 downto 0);
	     bitcount : std_logic_vector(3 downto 0);
	     -- Buffered outputs
	     txd : std_logic;
	     txempty : std_logic;
    end record;

    type recieve_reg_type is record
	     -- Internal state variables
	     state : state_type;
	     data : std_logic_vector(7 downto 0);
	     mr_factor : std_logic_vector(11 downto 0);
	     bitcount : std_logic_vector(3 downto 0);
	     -- Buffered outputs
	     rxready : std_logic;
	     rcv_ack : std_logic;
	     dataport : std_logic_vector(7 downto 0);
    end record;
    
    signal toext : send_reg_type;
    signal toext_in : send_reg_type;
    signal fromext : recieve_reg_type;
    signal fromext_in : recieve_reg_type;
    
begin

    ---------------------------------------------------------------------------
    -- THIS PROCESS WILL BUFFER ALL SIGNALS COMING INTO THIS ARCHITECTURE
    ---------------------------------------------------------------------------
    
    mr_reg : process(clk, toext_in, fromext_in)
    begin
	if (rising_edge(clk)) then
	    -- Stuff to send to outer device, first copy output from
	    -- previous run.
	    txd <= toext_in.txd;
	    uarto.txempty <= toext_in.txempty;
	    toext <= toext_in;
	    -- Stuff to recieve from outer device, copy from previous
	    uarto.rxready <= fromext_in.rxready;
	    uarto.data <= fromext_in.dataport;
	    fromext <= fromext_in;
	end if;
    end process;

    
    ---------------------------------------------------------------------------
    -- THIS PROCESS DEALS WITH SENDING DATA TO EXTERNAL UNIT
    ---------------------------------------------------------------------------

    -- fixat... underbart att lyssna p radio.c64.org med yodelking
    -- som skivvndare

    sender : process(rst, toext, uarti)
	variable t : send_reg_type;
    begin
	t := toext;
	t.mr_factor := t.mr_factor - 1;
	case t.state is
	    when IDLE =>
		if (uarti.send = '1') then
		    -- Sample and hold input line
		    t.data := uarti.data;
		    -- Tell controller to take it easy
		    t.txempty := '0';
		    -- Switch state.
		    t.state := STARTBIT;
		    t.mr_factor := MR_CLOCKFACTOR_OUT;
		    -- Here comes the startbit
		    t.txd := '0';
		    -- print("BEGARAN OM SEND STARTBIT");
		else
		    -- print("IDLE");
		end if;
	    when STARTBIT =>
		if (t.mr_factor = 0) then
		    t.state := DATA;
		    t.bitcount := "0111";  -- Detta r de 7 resterande bitarna
		    t.txd := t.data(0);  -- Frsta databiten
		    t.data := shr(t.data, "1");  -- Skifta
		    -- print("SENT STARTBIT");
		end if;
	    when DATA =>
		if (t.mr_factor = 0) then
		    t.bitcount := t.bitcount - 1;
		    t.txd := t.data(0);  -- Skicka en bit
		    t.data := shr(t.data, "1");  -- Skifta
		    if (t.bitcount = 0) then
			-- Switch state
			t.state := STOPBIT;
			-- This is to count the stop bit
			t.bitcount := "0010";
			-- print("REQUEST STOPBIT");
		    end if;
		end if;
	    when STOPBIT =>
		if (t.mr_factor = 0) then
		    -- Send stop bit
		    t.txd := '1';
		    t.bitcount := t.bitcount - 1;
		    if (t.bitcount = 0) then
			-- Tell the using entity that we are ready sending
			-- this byte.
			t.txempty := '1';
			t.state := IDLE;
			-- print("SENT STOPBIT");
		    end if;
		end if;
	    when ACK =>
		-- This state is not used when sending.
		t.state := IDLE;
	end case;

	if (t.mr_factor = 0) then
	    t.mr_factor := MR_CLOCKFACTOR_OUT;
	end if;

	-- First, if we want to reset the whole shebang, active low
	if (rst = '0') then
	    t.state := IDLE;
	    -- The send queue IS empty.
	    t.txempty := '1';
	    t.mr_factor := MR_CLOCKFACTOR_OUT;
	    -- The txd line should be 1 when not used (=stopbit)
	    t.txd := '1';
	end if;

	-- Our only outsignal.
	toext_in <= t;
	
end process;

    
     -----------------------------------------------------------------------
     --	THIS PROCESS DEALS WITH RECIEVING DATA FROM EXTERNAL UNIT
     -----------------------------------------------------------------------

    reciever : process(rst, fromext, rxd, uarti)
	variable r : recieve_reg_type;
    begin
	r := fromext;
	r.mr_factor := r.mr_factor - 1;
	case r.state is
	    when IDLE =>
		-- React on start bit
		if (rxd = '0') then
		    -- Count half mrFactor to reach middle of start bit
		    r.mr_factor := shr(MR_CLOCKFACTOR_IN, "1");
		    r.state := STARTBIT;
		    -- print("SWITCH TO STARTBIT");
		end if;
	    when STARTBIT =>
		if (r.mr_factor = 0) then
		    -- We are now in the middle of the possible start bit
		    if (rxd = '0') then
			r.bitcount := "1000";
			-- Toto! We're not in Kansas anymore!
			r.rxready := '0';
			r.state := DATA;
			-- print("SWITCH TO DATA");
		    else
			r.state := IDLE;
		    end if;
		end if;
	    when DATA =>
		if (r.mr_factor = 0) then
		    -- Shift in the happy databit
		    r.data := shr(r.data, "1");
		    r.data(7) := rxd;
		    -- Count all 8 bits
		    r.bitcount := r.bitcount - 1;
		    if (r.bitcount = 0) then
			r.state := STOPBIT;
			-- print("SWITCH TO STOPBIT");
		    end if;
		end if;
	    when STOPBIT =>
		if (r.mr_factor = 0) then
		    if (rxd = '1') then
			-- Update and hold data value on uart output
			r.dataport := r.data;
			r.rxready := '1';
			-- Go to ACK
			r.state := ACK;
			-- print("SWITCH BACK TO ACK");
		    end if;
		end if;
	    when ACK =>
		if (uarti.ack_rcv = '1') then
		    r.state := IDLE;
		    -- Not ready anymore.
		    r.rxready := '0';
		    -- print("SWITCH BACK TO IDLE");
		elsif (r.mr_factor = 0) then
		    -- TIMEOUT. Go to IDLE.
		    r.state := IDLE;
		    r.rxready := '0';
		    -- print("SWITCH BACK TO IDLE");
		end if;
	end case;
	if (r.mr_factor = 0) then
	    r.mr_factor := MR_CLOCKFACTOR_IN;
	end if;

	-- First, if we want to reset the whole shebang.
	if (rst = '0' ) then
	    r.state := IDLE;
	    -- And the recieve queue IS empty, nothing to recieve here.
	    r.rxready := '0';
	    r.mr_factor := MR_CLOCKFACTOR_IN;
	    -- Why not zero the output register on startup.
	    r.dataport := "00000000";
	end if;
	fromext_in <= r;

    end process;
end;

