library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_signed.all;
use IEEE.std_logic_arith.all;
use std.textio.all;
use work.tbcomp.all;
use work.memcomp.all;
use work.controller_types.all;
use work.uart_lcd_pck.all;

entity tb_controller is
end;

architecture behavioral of tb_controller is

  constant FLASH_FILE_SYSTEM : string := "debug/flash.bit";                -- flash memory contents
  constant MP3_FILENAME      : string := "debug/song.dat";                 -- data sent to the MP3 chip
  constant LCD_FILENAME      : string := "debug/lcd.txt";                  -- data received by the LCD
  constant DAC_FILENAME      : string := "debug/dac.txt";                  -- data received by the DAC

  component controller_top
    port( clk         : in  std_logic;                                      -- system clock 12MHz
          rst         : in  std_logic;                                      -- system reset
          flash_addr  : out std_logic_vector(FLASH_ADDR_BITS-1 downto 0);   -- address of the external FLASH
          flash_data  : in  std_logic_vector(FLASH_DATA_BITS-1 downto 0);   -- data from the external FLASH
	  -- flash_wdata : out std_logic_vector(FLASH_DATA_BITS-1 downto 0);	  -- data to the external FLASH
	  -- flash_en    : out std_logic;
	  -- flash_we    : out std_logic;
	  -- flash_oe    : out std_logic;
          button_play : in  std_logic;                                      -- play
          button_stop : in  std_logic;                                      -- stop
          button_prev : in  std_logic;                                      -- previous song
          button_next : in  std_logic;                                      -- next song
          mp3_rst     : out std_logic;                                      -- to MP3 chip
          mp3_data    : out std_logic;                                      -- to MP3 chip
          mp3_valid   : out std_logic;                                      -- to MP3 chip
          mp3_req     : in  std_logic;                                      -- from MP3 chip
          spi_cs      : out std_logic;                                      -- D/A control (data)
          spi_sclk    : out std_logic;                                      -- D/A control (clk)
          spi_din     : out std_logic;                                      -- D/A control (clk)
          rxd         : in  std_logic;                                      -- serial input (from PC)
          txd         : out std_logic                                       -- serial output (display)
        );
  end component;

  procedure print(s : string) is
    variable L1 : line;
  begin
    L1:= new string'(s);
    writeline(output,L1);
  end;
  
  signal clk, rst, gnd, vdd                     : std_logic;    
  signal mp3_rst, mp3_data, mp3_valid, mp3_req  : std_logic;
  signal spi_cs, spi_din, spi_sclk              : std_logic;
  signal rxd, txd                               : std_logic; 

  signal flash_addr   : std_logic_vector(FLASH_ADDR_BITS-1 downto 0);
  signal flash_data   : std_logic_vector(FLASH_DATA_BITS-1 downto 0);

  signal button_play  : std_logic;
  signal button_stop  : std_logic;
  signal button_prev  : std_logic;
  signal button_next  : std_logic;

begin

  ----------------------------
  -- Controller
  ----------------------------

  ctrl: controller_top
    port map ( clk         => clk,
               rst         => rst,
               flash_addr  => flash_addr,
               flash_data  => flash_data,
               button_play => button_play,
               button_stop => button_stop,
               button_prev => button_prev,
               button_next => button_next,
               mp3_rst     => mp3_rst,
               mp3_data    => mp3_data,
               mp3_valid   => mp3_valid,
               mp3_req     => mp3_req,
               spi_cs      => spi_cs,
               spi_sclk    => spi_sclk,
               spi_din     => spi_din,
               rxd         => rxd,
               txd         => txd
             );


  ----------------------------
  -- Flash
  ----------------------------

  mem: flash
    generic map ( ADDR_BITS => FLASH_ADDR_BITS,
                  DATA_BITS => FLASH_DATA_BITS,
                  ROM_FILE  => FLASH_FILE_SYSTEM )
    port map ( clk  => clk,
               addr => flash_addr,
               do   => flash_data
             );

  ----------------------------
  -- LCD (RS232 interface)
  ----------------------------

  lcd: uart_lcd
    generic map ( MASTER_CLOCK => SYSTEM_CLOCK,
                  BAUD_RATE => BAUD_RATE_LCD,
                  LCD_FILENAME => LCD_FILENAME
                )
    port map ( clk   => clk,
               rst   => rst,
               rxd   => txd
             );

  ----------------------------
  -- Clock and Reset
  ----------------------------

  clkgen: clock_generator
    generic map( 83 ns )
    port map ( CLK => CLK );

  rstgen: reset_generator
    generic map( 450 ns )
    port map ( RST => RST );

  gnd <= '0';
  vdd <= '1';  

  process
  begin
    print(".");        
    print("------------------------------------------------------------");
    print("Controller settings");
    print("------------------------------------------------------------");
    print(".");        
    print("Target:     " & target_type'IMAGE(TARGET_TECH));        
    print(".");        
    print("Flash file system:     " & FLASH_FILE_SYSTEM);        
    print("MP3 chip data output:     " & MP3_FILENAME);        
    print("LCD data receiver:     " & LCD_FILENAME);        
    print("DAC data receiver:     " & DAC_FILENAME);        
    print(".");        
    print("System clock:     " & INTEGER'IMAGE(SYSTEM_CLOCK));        
    print("LCD baud rate:     " & INTEGER'IMAGE(BAUD_RATE_LCD));        
    print("PC baud rate:     " & INTEGER'IMAGE(BAUD_RATE_PC));        
    print(".");        
    print("------------------------------------------------------------");
    wait;
  end process;

  ----------------------------
  -- MP3 chip emulator
  ----------------------------

  process
    variable sample     : std_logic_vector(FLASH_DATA_BITS-1 downto 0);
    variable counter, delay    : integer;
    file     mp3_chip_out : TEXT OPEN WRITE_MODE IS MP3_FILENAME;
    variable value      : line;
  begin
    mp3_req <= '0';
    counter := FLASH_DATA_BITS - 1;

    print("MP3 chip is waiting for reset pulse");
    wait until mp3_rst = '1';
    print("MP3 chip ready, asserting request");
                          
    loop
      case counter is
        when 0 => delay := 2; 
        when 1 => delay := 0; 
        when 2 => delay := 1; 
        when 3 => delay := 3; 
        when others => delay := 2;
      end case;

      if (delay > 0) then
        for i in 1 to delay loop
          mp3_req <= '0';                     -- stop request
          wait until clk = '1';
        end loop;
      end if;

      mp3_req <= '1';                         -- request data
      wait until clk = '1';                   -- wait for clock
      if (mp3_valid = '0') then
        wait until mp3_valid = '1';           -- wait for valid data
      end if;
      wait until clk = '0';                   -- wait for clock
      sample(counter) := mp3_data;            -- store sample
                                              
      if (counter = 0) then                   -- received one word
        write(value, to_bitvector(sample));   -- write to file
        writeline(mp3_chip_out, value);               
        counter := FLASH_DATA_BITS;   
      end if;
      counter := counter - 1;
    end loop;
  end process;

  ----------------------------
  -- D/A SPI emulator
  ----------------------------

  process
    variable sreg    : std_logic_vector(15 downto 0);
    file     dac_out : TEXT OPEN WRITE_MODE IS DAC_FILENAME;
    variable value   : line;
    variable addr    : string(1 to 9) := "Address: ";
    variable data    : string(1 to 9) := "   Data: ";
  begin
    sreg := (others => '0');    

    wait until rst = '1';
    
    loop
      wait until spi_cs = '0';
      loop
        wait until spi_sclk = '1' or spi_cs = '1';
        if (spi_cs = '1') then
          write(value, addr);
          write(value, to_bitvector(sreg(15 downto 8)));
          write(value, data);
          write(value, to_bitvector(sreg(7 downto 0)));
          writeline(dac_out, value);               
          exit;
        end if;
        sreg := sreg(14 downto 0) & spi_din;
      end loop;
    end loop;
  end process;

end;


-- uncomment the lines below to do post simulation

configuration cfg_controller of tb_controller is
    for behavioral
	for ctrl: controller_top
	    use entity work.controller_top(SYN_structural);
	end for;
    end for;
end;

