library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_signed.all;
use IEEE.std_logic_arith.all;
use std.textio.all;
use work.tbcomp.all;
use work.memcomp.all;
use work.controller_types.all;
use work.flashctl_pck.all;

entity tb_flashctl is
end;

architecture behavioral of tb_flashctl is
    
  constant FLASH_FILE_SYSTEM : string := "debug/flash.bit";                -- flash memory contents
  constant MP3_FILENAME      : string := "debug/song.dat";                 -- data sent to the MP3 chip

  signal clk, cpu_clk, rst, gnd, vdd                     : std_logic;
  -- Memory stuff
  signal flash_rom_data_out : std_logic_vector(7 downto 0);
  signal flash_rom_re : std_logic;
  signal flash_rom_address_in : std_logic_vector(15 downto 0);
  -- Flash controller play registers
  signal flash_play_address : std_logic_vector (31 downto 0);
  signal flash_stop_address : std_logic_vector (31 downto 0);
  signal flash_play : std_logic;
  signal flash_stop : std_logic;
  signal flash_stopped : std_logic;
  -- Connext to flask memory
  signal flashi : flash_out_type;
  signal flasho : flash_in_type;
  -- Interface to MP3 chip
  signal mp3i : mp3_out_type;
  signal mp3o : mp3_in_type;

  -- Used to count cycles
  signal mr_counter : std_logic_vector(7 downto 0);

  -- MP3 Chip simulator
  signal mp3_rst, mp3_data, mp3_valid, mp3_req  : std_logic;
  
  procedure print(s : string) is
    variable L1 : line;
  begin
    L1:= new string'(s);
    writeline(output,L1);
  end;
  
begin

  mp3_rst <= mp3o.rst;
  mp3_data <= mp3o.data;
  mp3_valid <= mp3o.valid;
  mp3_req <= mp3i.req;
    
  mem: flash
    generic map ( ADDR_BITS => FLASH_ADDR_BITS,
                  DATA_BITS => FLASH_DATA_BITS,
                  ROM_FILE  => FLASH_FILE_SYSTEM )
    port map ( clk  => clk,
               addr => flasho.addr,
               do   => flashi.rdata
             );

  ----------------------------
  -- FLASH controller
  ----------------------------

  mr_flash : flashctl
      port map (clk => clk,
		rst => rst,
		cpu_clk => cpu_clk,
		flashi => flashi,
		flasho => flasho,
		flash_rom_data_out => flash_rom_data_out,
		flash_rom_re => flash_rom_re,
		flash_rom_address_in => flash_rom_address_in,
		play_start_address => flash_play_address,
		play_end_address => flash_stop_address,
		play => flash_play,
		stop => flash_stop,
		stopped => flash_stopped,
		mp3i => mp3i,
		mp3o => mp3o
	    );

  
  ----------------------------
  -- Clock and Reset
  ----------------------------

  clkgen: clock_generator
    generic map( 83 ns )
    port map ( CLK => CLK );

  rstgen: reset_generator
    generic map( 450 ns )
    port map ( RST => RST );

  gnd <= '0';
  vdd <= '1';

  bar : process(clk, rst)
  begin
      if (rst = '0') then
	  -- flasho.addr <= "0000000000000000";
	  mr_counter <= "00000000";
	  flash_rom_re <= '0';
	  flash_play <= '0';
      elsif rising_edge(clk) then
	  mr_counter <= mr_counter + 1;
	  -- Test to read from rom address $FFFD
	  -- cpu_addr <= "1111111111111101";
	  -- cpu_re <= '1';
	  case mr_counter is
	      when "00000001" =>
		  -- Test to read from address $00000000
		  -- flash_rom_address_in <= "0000000000000000";
		  flash_rom_address_in <= "1111111111111100";
		  flash_rom_re <= '1';
	      when "00000010" =>
		  flash_rom_re <= '0';
	      when "00000101" =>
		  flash_rom_address_in <= "1111111111111101";
		  flash_rom_re <= '1';
	      when "00000110" =>
		  flash_rom_re <= '0';
-- 	      when "00000100" =>
-- 		  -- Spela "wiggen"!
-- 		  flash_play_address <= "00000000000000010000000000000000";
-- 		  flash_stop_address <= "00000000000000010000000000000000";
-- 		  flash_play <= '1';
-- 	      when "00000101" =>
-- 		  flash_play <= '0';
	      when others =>
	  end case;
      end if;
  end process;

  ----------------------------
  -- MP3 chip emulator
  ----------------------------
  
  process
    variable sample     : std_logic_vector(FLASH_DATA_BITS-1 downto 0);
    variable counter, delay    : integer;
    file     mp3_chip_out : TEXT OPEN WRITE_MODE IS MP3_FILENAME;
    variable value      : line;
  begin
    mp3_req <= '0';
    counter := FLASH_DATA_BITS - 1;

    print("MP3 chip is waiting for reset pulse");
    wait until mp3_rst = '1';
    print("MP3 chip ready, asserting request");
                          
    loop
      case counter is
        when 0 => delay := 2; 
        when 1 => delay := 0; 
        when 2 => delay := 1; 
        when 3 => delay := 3; 
        when others => delay := 2;
      end case;

      if (delay > 0) then
        for i in 1 to delay loop
          mp3_req <= '0';                     -- stop request
          wait until clk = '1';
        end loop;
      end if;

      mp3_req <= '1';                         -- request data
      wait until clk = '1';                   -- wait for clock
      if (mp3_valid = '0') then
        wait until mp3_valid = '1';           -- wait for valid data
      end if;
      wait until clk = '0';                   -- wait for clock
      sample(counter) := mp3_data;            -- store sample
                                              
      if (counter = 0) then                   -- received one word
        write(value, to_bitvector(sample));   -- write to file
        writeline(mp3_chip_out, value);               
        counter := FLASH_DATA_BITS;   
      end if;
      counter := counter - 1;
    end loop;
  end process;

  
end;







