library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_signed.all;
use IEEE.std_logic_arith.all;
use std.textio.all;
use work.tbcomp.all;
use work.memcomp.all;
use work.mmu_pck.all;
use work.uart_pck.all;
use work.controller_types.all;
use work.ram.all;

entity tb_mmu is
end;

architecture behavioral of tb_mmu is

  constant FLASH_FILE_SYSTEM : string := "debug/flash.bit";                -- flash memory contents
  constant MP3_FILENAME      : string := "debug/song.dat";                 -- data sent to the MP3 chip

  procedure print(s : string) is
    variable L1 : line;
  begin
    L1:= new string'(s);
    writeline(output,L1);
  end;
  
  signal clk, halt, rst, gnd, vdd                     : std_logic;
  -- CPU signals
  signal cpu_addr : std_logic_vector (15 downto 0) := "0000000000000000";
  signal cpu_in   : std_logic_vector (7 downto 0) := "00000000";
  signal cpu_out  : std_logic_vector (7 downto 0) := "00000000";
  signal cpu_we   : std_logic := '0';
  signal cpu_re   : std_logic := '0';
  -- RAM circuit
  signal ram_data_in : std_logic_vector (7 downto 0);
  signal ram_data_out : std_logic_vector (7 downto 0);
  signal ram_we : std_logic := '0';
  -- LCD ports and stuff
  signal uarti    : uart_ctrl_in_type;
  signal uarto    : uart_ctrl_out_type;
  -- Used to count cycles
  signal mr_counter : std_logic_vector(7 downto 0);
  -- Buttons
  signal button_play : std_logic;
  signal button_stop : std_logic;
  signal button_prev : std_logic;
  signal button_next : std_logic;
  -- Clock
  signal clock_rst     : std_logic;
  signal clock_hold    : std_logic;
  signal clock_read    : std_logic;
  signal clock_seconds : std_logic_vector(7 downto 0);
  signal clock_minutes : std_logic_vector(7 downto 0);
  signal clock_hours   : std_logic_vector(7 downto 0);
  -- Connext to flask memory
  signal flashi : flash_out_type;
  signal flasho : flash_in_type;
  -- Interface to MP3 chip
  signal mp3i : mp3_out_type;
  signal mp3o : mp3_in_type;
  -- MP3 Chip simulator
  signal mp3_rst, mp3_data, mp3_valid, mp3_req  : std_logic;
  -- DAC
  signal dac_data_byteh  : std_logic_vector(7 downto 0);
  signal dac_data_bytel  : std_logic_vector(7 downto 0);
  signal dac_send_data   : std_logic;   
  signal dac_data_req    : std_logic;   

  signal ram_we_inv : std_logic;
begin

  mp3_rst <= mp3o.rst;
  mp3_data <= mp3o.data;
  mp3_valid <= mp3o.valid;
  mp3_req <= mp3i.req;
    
  mem: flash
    generic map ( ADDR_BITS => FLASH_ADDR_BITS,
                  DATA_BITS => FLASH_DATA_BITS,
                  ROM_FILE  => FLASH_FILE_SYSTEM )
    port map ( clk  => clk,
               addr => flasho.addr,
               do   => flashi.rdata
             );
    
  ----------------------------
  -- UART
  ----------------------------

  mr_mmu : mmu
      port map (clk => clk,
		halt => halt,
		rst => rst,
		cpu_addr => cpu_addr,
                cpu_in => cpu_in,
                cpu_out => cpu_out,
                cpu_we => cpu_we,
                cpu_re => cpu_re,
		ram_data_in => ram_data_in,
		ram_data_out => ram_data_out,
		ram_we => ram_we,
		uarti => uarti,
		uarto => uarto,
		flasho => flasho,
		flashi => flashi,
		button_play => button_play,
		button_stop => button_stop,
		button_prev => button_prev,
		button_next => button_next,
		clock_rst => clock_rst,
		clock_hold => clock_hold,
		clock_read => clock_read,
		clock_seconds => clock_seconds,
		clock_minutes => clock_minutes,
		clock_hours => clock_hours,
		mp3i => mp3i,
		mp3o => mp3o,
		dac_data_byteh => dac_data_byteh,
		dac_data_bytel => dac_data_bytel,
		dac_send_data => dac_send_data,
		dac_data_req => dac_data_req
		);

  -- LssasRAM
  mr_ram : mp3ram
      generic map (BYTES => 1024)
      -- Remember to map adress line immediately to ROM
      port map ( addr => cpu_addr(9 downto 0),
		 clk => clk,
		 data_in => ram_data_in,
		 data_out => ram_data_out,
		 we => ram_we);


  ----------------------------
  -- Clock and Reset
  ----------------------------

  clkgen: clock_generator
    generic map( 83 ns )
    port map ( CLK => CLK );

  rstgen: reset_generator
    generic map( 450 ns )
    port map ( RST => RST );

  gnd <= '0';
  vdd <= '1';

  bar : process(clk, rst)
      variable text_line : line;
      variable indata : integer;
  begin
      if (rst = '0') then
	  cpu_addr <= "0000000000000000";
	  cpu_re <= '0';
	  mr_counter <= "00000000";
      elsif rising_edge(clk) then
	  mr_counter <= mr_counter + 1;
	  -- Test to read from rom address $FFFD
	  -- cpu_addr <= "1111111111111101";
	  -- cpu_re <= '1';
	  case mr_counter is

-- 	      when "00000001" =>
-- 		  -- Test to write to port $8002
-- 		  cpu_addr <= "1000000000000010";
-- 		  cpu_out <= "10101010";
-- 		  cpu_we <= '1';
-- 	      when "00000010" =>
-- 		  cpu_we <= '0';
-- 	      when "00000100" =>
-- 		  -- Test to write to port $8001
-- 		  cpu_addr <= "1000000000000001";
-- 		  cpu_out <= "10000000";
-- 		  cpu_we <= '1';
-- 	      when "00000101" =>
-- 		  cpu_we <= '0';

	      
-- 	      when "00000001" =>
-- 		  -- Test to write to RAM $0002
-- 		  cpu_addr <= "0000000000000010";
-- 		  cpu_out <= "10101010";
-- 		  cpu_we <= '1';
-- 	      when "00000010" =>
-- 		  cpu_addr <= "0000000000000000";
-- 		  cpu_out <= "UUUUUUUU";
-- 		  cpu_we <= '0';
-- 	      when "00000100" =>
-- 		  -- Test to read from RAM $0002
-- 		  cpu_addr <= "0000000000000010";
-- 		  cpu_re <= '1';
-- 	      when "00000101" =>
-- 		  cpu_re <= '0';
		  
-- 	      when "00001001" =>
-- 		  -- Test to write to RAM $0003
-- 		  cpu_addr <= "0000000000000011";
-- 		  cpu_out <= "01010101";
-- 		  cpu_we <= '1';
-- 	      when "00001010" =>
-- 		  cpu_addr <= "0000000000000000";
-- 		  cpu_out <= "UUUUUUUU";
-- 		  cpu_we <= '0';
-- 	      when "00001100" =>
-- 		  -- Test to read from RAM $0003
-- 		  cpu_addr <= "0000000000000011";
-- 		  cpu_re <= '1';
-- 	      when "00001101" =>
-- 		  cpu_addr <= "0000000000000000";
-- 		  cpu_re <= '0';

-- 	      when "00001111" =>
-- 		  -- Test to read from RAM $0001
-- 		  cpu_addr <= "0000000000000001";
-- 		  cpu_re <= '1';
-- 	      when "00010000" =>
-- 		  -- Test to read from RAM $0002
-- 		  cpu_addr <= "0000000000000010";
-- 		  cpu_re <= '1';
-- 	      when "00010001" =>
-- 		  cpu_re <= '0';		  

	      when "00000001" =>
		  -- Read from PROM $f000 $f001
		  cpu_addr <= "1111000000000000";
		  cpu_re <= '1';
	      when "00001000" =>
		  cpu_addr <= "1111000000000001";
	      when "00001111" =>
		  cpu_re <= '0';

-- 	      when "00000001" =>
-- 		  -- Read from PROM $Fffc $Fffd
-- 		  cpu_addr <= "1111111111111100";
-- 		  cpu_re <= '1';
-- 	      when "00001000" =>
-- 		  cpu_addr <= "1111111111111101";
-- 	      when "00001111" =>
-- 		  cpu_re <= '0';		  

	      when others =>
	  end case;
      end if;
  end process;

  ram_we_inv <= not ram_we;
end;


