---------------------------------------------------
--
-- Simple UART controller for arbitrary baud rates
--                    
-- Receiver only
--
-- Copyright 2003 Thomas Lenart tlt@es.lth.se
--
---------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

package uart_lcd_pck is
    
    component uart_lcd
	generic ( MASTER_CLOCK : integer;
		  BAUD_RATE: integer;
		  LCD_FILENAME : string
		  );
	port ( clk   : in  std_logic;
	       rst   : in  std_logic;
	       rxd   : in  std_logic
	       );
    end component;
    
end;


-------------------------------------------------------------------------------
-- Entity stuff
-------------------------------------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use std.textio.all;
use work.uart_lcd_pck.all;

entity uart_lcd is
  generic ( MASTER_CLOCK : integer := 12_000_000;
            BAUD_RATE    : integer := 9600;
            LCD_FILENAME : string  := "lcd.txt"
          );
  port ( clk   : in  std_logic;
         rst   : in  std_logic;
         rxd   : in  std_logic
       );
end;

architecture behav of uart_lcd is

  constant SCALER : std_logic_vector(11 downto 0) := CONV_STD_LOGIC_VECTOR(MASTER_CLOCK / (8*BAUD_RATE), 12);

  type state_type is ( IDLE, STARTBIT, DATA, STOPBIT );

  type reg_type is record
    scaler  : std_logic_vector(11 downto 0);
    rxstate : state_type;
    rxshift : std_logic_vector(7 downto 0);
    rxdata  : std_logic_vector(7 downto 0);
    tick    : std_logic;
    rxtick  : std_logic;
    rxclk   : std_logic_vector(2 downto 0);
    rxready : std_logic;    
    rxd     : std_logic;    
  end record;

  signal r, rin : reg_type;

begin

  process(r, rst, rxd)
    variable v : reg_type;
  begin
    v := r;
    v.tick    := '0';
    v.rxtick  := '0';
    v.rxready := '0';

-- RX counters
------------------------------------------------------------------------------------------------------------------------

-- pragma translate_off
    if not is_x(r.scaler) then		-- avoid warnings at reset time
-- pragma translate_on
      v.scaler := r.scaler - 1;
-- pragma translate_off
    end if;
-- pragma translate_on
    v.tick := v.scaler(11) and not r.scaler(11);
    if (v.tick = '1') then 
      v.scaler := SCALER;
    end if;
    
    if (r.tick = '1') then
-- pragma translate_off
      if not is_x(r.rxclk) then		-- avoid warnings at reset time
-- pragma translate_on
        v.rxclk := r.rxclk + 1;
-- pragma translate_off
      end if;
-- pragma translate_on
      v.rxtick := r.rxclk(2) and not v.rxclk(2);
    end if;

-- Receiver
------------------------------------------------------------------------------------------------------------------------
    
    case r.rxstate is
      when IDLE     => if (r.rxd = '0') then                            -- wait for leading zero (startbit)
                         v.rxstate := STARTBIT;                         
                         v.rxclk := "100";                              -- delay 4 cycles first time
                         v.rxshift := (others => '1');                  -- fill shift register with ones to locate the startbit later
                         v.rxtick := '0';                               -- avoid previously generated ticks
                       end if;
      when STARTBIT => if (r.rxtick = '1') then                         -- wait for rxtick that is 4 ticks
                         if (r.rxd = '0') then                          -- if this is still a start bit
                           v.rxshift := '0' & r.rxshift(7 downto 1);    -- put start bit in shift register
                           v.rxstate := DATA;                           -- start receiving data
                         else
                           v.rxstate := IDLE;                           -- abort due to start bit error
                         end if;
                       end if;
      when DATA     => if (r.rxtick = '1') then                         -- wait for rxtick that is 8 ticks
                         v.rxshift := r.rxd & r.rxshift(7 downto 1);    -- put data in the shift register
                         if (r.rxshift(0) = '0') then                   -- if 8 bits has been received
                           v.rxstate := STOPBIT;                        -- wait for stop bit
                         end if;                           
                       end if;
      when STOPBIT  => if (r.rxtick = '1') then
--                         if (r.rxd = '1') then
                           v.rxstate := IDLE;
                           v.rxdata  := r.rxshift;
                           v.rxready := '1';
--                         end if;
                       end if;
    end case;

    if (rst = '0') then
      v.scaler  := (others => '0');
      v.rxstate := IDLE;
    end if;
    
    v.rxd := rxd;

    rin <= v;
    
  end process;

  process(clk, rin)
  begin
    if rising_edge(clk) then
      r <= rin;
    end if;
  end process;

  process
    file f : TEXT is out LCD_FILENAME;
    variable lcd_line : line; 
    variable ch : integer;
  begin
    loop
      wait until clk='1';
      if (r.rxready = '1') then
        if (not is_x(r.rxdata)) then
--          ch := CHARACTER'VAL(CONV_INTEGER(r.rxdata));           
          ch := CONV_INTEGER(r.rxdata);           
          write(lcd_line, ch);
          writeline(f, lcd_line);        
        end if;
      end if;
    end loop;
  end process;
                     
end;
